import onnx

from ..__version import __version__
from .infer_type import *

onnx_dtype = {
    0: 'UNDEFINED',    1: 'float',    2: 'uint8',    3: 'int8',    4: 'uint16',
    5: 'int16',    6: 'int32',    7: 'int64',    8: 'str',    9: 'bool',    10: 'float16',
    11: 'double',    12: 'uint32',    13: 'uint64',    14: 'complex64',    15: 'complex128',
    16: 'bfloat16'
}

only_need_input0_op_list = ['Reshape']


def get_node_id(node, nodes):
    for k in range(len(nodes)):
        if node == nodes[k]:
            return k


def get_index_by_name(self, name):
    for i in range(len(self.model.graph.node)):
        if self.model.graph.node[i].name == name:
            return i


def get_node_by_index(model, index):
    return model.graph.node[index]


def insert_op_before(model, node_input, node_output, target_node_index, input_i, *args, **kwargs):
    '''
    op_name
    weight_dict
    attr_dict
    '''

    model.graph.node[target_node_index].input[input_i] = "Dequant_" + \
        str(node_input)+"_"+str(node_output)
    new_dequant_node = onnx.helper.make_node(
        'Dequant',
        inputs=[node_input],
        outputs=["Dequant_"+str(node_input)+"_"+str(node_output)],
        domain="thinker",
        **kwargs['attr_dict']
    )

    # set target_node input to new node outputs

    model.graph.node.insert(target_node_index, new_dequant_node)


def add_dequant_layer(model, target_node_index, saved_node, params, input_i):

    saved_scale_o = float(params[0])

    weight_dict = {
    }
    attr_dict = {
        "scale_o": saved_scale_o,
    }

    insert_op_before(
        model,
        node_input=saved_node.input[input_i],
        node_output=saved_node.output[0],
        target_node_index=target_node_index,
        input_i=input_i,
        weight_dict=weight_dict,
        attr_dict=attr_dict
    )


def insert_op_after(model, node_input, node_output, target_node_index, output_i, fixed_input, *args, **kwargs):
    '''
    op_name
    weight_dict
    attr_dict
    '''
    for each_node in model.graph.node:
        for input_idx in range(len(each_node.input)):
            if each_node.input[input_idx] == fixed_input:
                each_node.input[input_idx] = "Dequant_" + \
                    str(node_input)+"_"+str(node_output)

    model.graph.node[target_node_index].output[output_i] = "Dequant_" + \
        str(node_input)+"_"+str(node_output)
    new_dequant_node = onnx.helper.make_node(
        'Dequant',
        inputs=["Dequant_"+str(node_input)+"_"+str(node_output)],
        outputs=[node_input],
        domain="thinker",
        **kwargs['attr_dict']
    )

    # set target_node input to new node outputs
    model.graph.node.insert(target_node_index+1, new_dequant_node)


def add_dequant_layer_after(model, target_node_index, saved_node, params, output_i, fixed_input):

    saved_scale_o = float(params[0])

    weight_dict = {
    }
    attr_dict = {
        "scale_o": saved_scale_o,
    }

    insert_op_after(
        model,
        node_input=saved_node.output[output_i],
        node_output=saved_node.input[0],
        target_node_index=target_node_index,
        output_i=output_i,
        fixed_input=fixed_input,
        weight_dict=weight_dict,
        attr_dict=attr_dict
    )


def update_dequant_out(module, saved_node, other_params, output_i, fixed_input):

    origin_layer_index = get_node_id(saved_node, module.graph.node)

    add_dequant_layer_after(module, origin_layer_index,
                            saved_node, other_params, output_i, fixed_input)


def update_dequant(module, saved_node, other_params, input_i):

    origin_layer_index = get_node_id(saved_node, module.graph.node)

    add_dequant_layer(module, origin_layer_index,
                      saved_node, other_params, input_i)


def analyse_attribute(node, saved_scale_o):

    other_params = [saved_scale_o]

    for i in range(len(node.attribute)):

        if node.attribute[i].name == 'scale_o':
            other_params[0] = node.attribute[i].f
        if node.op_type == "BatchNorm2dInt":
            if node.attribute[i].name == 'scale_add_o':
                other_params[0] = node.attribute[i].f

    return other_params


def remove_identity_node(model, op_type="Identity"):
    graph_output_name = []
    for ii in model.graph.output:
        graph_output_name.append(ii.name)
    nodes = model.graph.node[::-1]
    for i, node in enumerate(nodes):
        if node.op_type == op_type:
            model.graph.node.remove(node)
            if node.output[0] in graph_output_name:
                for each in model.graph.node:
                    for idx in range(len(each.output)):
                        if each.output[idx] == node.input[0]:
                            each.output[idx] = node.output[0]
                for each in model.graph.node:
                    for idx in range(len(each.input)):
                        if each.input[idx] == node.input[0]:
                            each.input[idx] = node.output[0]

                for each in model.graph.node:
                    if each.op_type == "If":
                        for gi in range(len(each.attribute)):
                            for x_node in each.attribute[gi].g.node:
                                for idx in range(len(x_node.output)):
                                    if x_node.output[idx] == node.input[0]:
                                        x_node.output[idx] = node.output[0]
                                for idx in range(len(x_node.input)):
                                    if x_node.input[idx] == node.input[0]:
                                        x_node.input[idx] = node.output[0]

            else:
                for each in model.graph.node:
                    for idx in range(len(each.input)):
                        if each.input[idx] == node.output[0]:
                            each.input[idx] = node.input[0]
                for each in model.graph.node:
                    if each.op_type == "If":
                        for gi in range(len(each.attribute)):
                            for x_node in each.attribute[gi].g.node:
                                for idx in range(len(x_node.input)):
                                    if x_node.input[idx] == node.output[0]:
                                        x_node.input[idx] = node.input[0]

    return model


def strategy_list_process(strategy_type_list):
    strategy_list_input = {}
    strategy_list_output = {}
    for k, v in enumerate(strategy_type_list):

        input_temp = []
        output_temp = []
        for ele in strategy_type_list[v]:
            input_temp.append(ele[0])
            output_temp.append(ele[1])
        strategy_list_input[v] = input_temp
        strategy_list_output[v] = output_temp
    return strategy_list_input, strategy_list_output


def add_node_name(model):
    for node in model.graph.node:
        node.name = node.name+str(node.op_type)+'_'+str(node.output[0])
    return model


def infer_type_linger(model, is_change_in_out_type):
    multi_input_non_quant_op_list = ['Concat', 'Mul', 'Add', 'Sub', 'Div', 'Max']
    
    initializer_names = []
    for x in model.graph.initializer:
        initializer_names.append(x.name)
    
    input_idxs = []
    for idx, y in enumerate(model.graph.input):
        if y.name not in initializer_names:
            input_idxs.append(y.name)

    nodes = model.graph.node
    tensor_type_map = {}
    remove_node_all = []
    out_change_nodes = []
    output_idxs = {}
    for idx, y in enumerate(model.graph.output):
        output_idxs[y.name] = onnx_dtype[y.type.tensor_type.elem_type]
    for inp in model.graph.input:
        tensor_type_map[inp.name] = inp.type.tensor_type.elem_type
    for inp in model.graph.initializer:
        tensor_type_map[inp.name] = inp.data_type

    for i, node in enumerate(nodes):
        in_type = []
        
        try:
            if node.op_type == 'LSTM':
                in_type = [tensor_type_map[node.input[0]]]
            elif node.op_type == 'GRU':
                in_type = [tensor_type_map[node.input[0]],
                           tensor_type_map[node.input[1]], tensor_type_map[node.input[2]]]
            else:
                in_type = [tensor_type_map[inp] for inp in node.input]
            
            if node.op_type in op_map.keys():
                it = op_map[node.op_type](node)
                # consider  slice/split  no-quant float situation ,   split don't support intop
                if node.op_type == "Slice" or node.op_type == "Split":
                    for quant_node in nodes:
                        for iii in range(len(node.output)):
                            if quant_node.input == [node.output[iii]] and quant_node.op_type == "OnnxInferQuant":
                                if in_type[0] == 3:
                                    in_type[0] = 0
            else:
                print("Warning: InferType OP ", node.op_type,
                      " is not supported,this may cause error !")
                it = op_map['Others'](node)

            if node.op_type == 'If':
                tensor_type_map.update(it.infer_type(in_type, tensor_type_map))
            else:
                tensor_type_map.update(it.infer_type(in_type))
        except:
            flag = 1
            for node_input_i in node.input:  # does it exist Quant op
                if node_input_i in input_idxs:
                    for iid in model.graph.input:
                        if iid.name == node_input_i and node.op_type not in multi_input_non_quant_op_list:
                            iid.type.tensor_type.elem_type = 3
                            tensor_type_map[node_input_i] = 3
                            flag = 0

            if flag:
                idx = 0
                if len(node.input) >= 2:
                    if node.op_type == "LSTM":
                        if tensor_type_map[node.input[0]] == 3:
                            remove_node_all.append((i, node, 0))
                    elif node.op_type == "GRU":
                        for x in range(3):
                            if tensor_type_map[node.input[x]] == 3:
                                remove_node_all.append((i, node, x))
                    elif node.op_type == "Clip":
                        if tensor_type_map[node.input[0]] == 3:
                            remove_node_all.append((i, node, 0))
                    else:
                        for x in range(len(node.input)):
                            if tensor_type_map[node.input[x]] == 3:
                                remove_node_all.append((i, node, x))
                else:
                    remove_node_all.append((i, node, idx))
                for output in node.output:
                    tensor_type_map[output] = 1
            else:
                for output in node.output:
                    tensor_type_map[output] = 3

        if not is_change_in_out_type:
            for output_i in range(len(node.output)):
                if node.output[output_i] in output_idxs.keys():
                    if tensor_type_map[node.output[output_i]] in [3, 6]:
                        out_change_nodes.append(
                            (i, node, output_i, node.output[output_i]))
        else:
            for output_i in range(len(node.output)):
                if node.output[output_i] in output_idxs.keys():
                    for oid in model.graph.output:
                        if oid.name == node.output[output_i]:
                            if tensor_type_map[node.output[output_i]] != oid.type.tensor_type.elem_type:
                                oid.type.tensor_type.elem_type = tensor_type_map[node.output[output_i]]

    return remove_node_all, out_change_nodes


def seach_scale_o(model, op_name):
    def seach_before(model, id_):
        for node in model.graph.node:
            if id_ in node.output:
                for num in range(len(node.attribute)):
                    if node.op_type != "Dequant" and node.attribute[num].name == "scale_o":
                        scale_o = node.attribute[num].f
                        return scale_o
                    if node.op_type != "Dequant" and node.attribute[num].name == "scale_add_o":
                        scale_o = node.attribute[num].f
                        return scale_o
                try:
                    return seach_before(model, node.input[0])
                except:
                    return 1.0

    for node in model.graph.node:
        if node.name == op_name:
            scale_o = seach_before(model, node.input[0])
            if scale_o is None:
                scale_o = 1.0
            if scale_o != node.attribute[-1].f:
                print("Scale_o Warning: The Dequant node (" + op_name +
                      ") attribute scale_o has wrong risk, Check the exported onnx model! ")

            node.attribute[-1].f = scale_o


def check_scale_o(model, op_name):
    def seach_before(model, id_):
        for node in model.graph.node:
            if id_ in node.output:
                for num in range(len(node.attribute)):
                    if node.op_type != "Dequant" and node.attribute[num].name == "scale_o":
                        scale_o = node.attribute[num].f
                        return scale_o
                    if node.op_type != "Dequant" and node.attribute[num].name == "scale_add_o":
                        scale_o = node.attribute[num].f
                        return scale_o
                try:
                    return seach_before(model, node.input[0])
                except:
                    return 1.0

    for node in model.graph.node:
        if node.name == op_name:
            scale_o = seach_before(model, node.input[0])
            if scale_o is None:
                scale_o = 1.0
            if scale_o != node.attribute[-1].f:
                print("Scale_o Error: The Dequant node (" + op_name +
                      ") attribute scale_o has wrong risk, Check the exported onnx model! ")


def parser_dequant(module, is_change_in_out_type):
    scale_o_list = []
    for i, node in enumerate(module.graph.node):
        attr_nums = len(node.attribute)
        attr_names = []
        saved_scale_o = 1.0000000000
        for num in range(attr_nums):
            attr_names.append(node.attribute[num].name)
            if node.attribute[num].name == "scale_o" or node.attribute[num].name == "scale_add_o":
                saved_scale_o = node.attribute[num].f

        if "scale_o" not in attr_names and "scale_add_o" not in attr_names:
            if len(scale_o_list) == 0:
                scale_o_list.append(1)
            else:
                scale_o_list.append(scale_o_list[-1])
        else:
            scale_o_list.append(saved_scale_o)

    all_op_infertyped_dict = {}

    initializer_names = []
    for x in module.graph.initializer:
        initializer_names.append(x.name)

    input_idxs = []
    for y in module.graph.input:
        if y.name not in initializer_names:
            input_idxs.append(y.name)

    output_idxs = {}
    for y in module.graph.output:
        output_idxs[y.name] = onnx_dtype[y.type.tensor_type.elem_type]

    for input_name in input_idxs:
        all_op_infertyped_dict[str(input_name)] = "float"

    remove_node_all = []
    out_change_nodes = []

    remove_node_all, out_change_nodes = infer_type_linger(
        module, is_change_in_out_type)

    for i_node in remove_node_all[::-1]:
        i = i_node[0]
        node = i_node[1]
        input_i = i_node[2]

        other_params = []

        saved_node = node
        saved_scale_o = scale_o_list[i]

        other_params = analyse_attribute(node, saved_scale_o)

        update_dequant(module, saved_node, other_params, input_i)

    for o_node in out_change_nodes[::-1]:
        i = o_node[0]
        node = o_node[1]
        output_i = o_node[2]
        fixed_input = o_node[3]
        other_params = []

        saved_node = node
        saved_scale_o = scale_o_list[i]

        other_params = analyse_attribute(node, saved_scale_o)

        update_dequant_out(module, saved_node, other_params,
                           output_i, fixed_input)

    module = remove_identity_node(module)

    module = add_node_name(module)

    module.producer_name = "linger_" + __version__ + '_' + module.producer_name

    for i, node in enumerate(module.graph.node):
        if node.op_type == "Dequant":
            seach_scale_o(module, node.name)

    # check scale_o
    for i, node in enumerate(module.graph.node):
        if node.op_type == "Constant" and '.weight' in node.output[0]:
            check_scale_o(module, node.name)

    for i, ini in enumerate(module.graph.initializer):
        if '.weight' in ini.name or '.bias' in ini.name:
            for node in module.graph.node:
                if len(node.input) > 0 and node.input[0] == ini.name and node.op_type == "Dequant":
                    print("Scale_o Error: The Dequant node (" + ini.name +
                          ") attribute scale_o is wrong ! ")

    return module


__all__ = ['parser_dequant']



import os

import linger
import numpy as np
import torch
import torch.nn as nn

if not os.path.exists('data.ignore'):
    os.mkdir('data.ignore')

def test_wb_analyse():
    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.conv = nn.Conv2d(2, 2, kernel_size=3, stride=1,
                                  padding=1, bias=True)
            self.conv1 = nn.Conv2d(2, 2, kernel_size=3, stride=1,
                                   padding=1, bias=True)
            self.fc = nn.Linear(8, 100)

        def forward(self, x):

            x = self.conv(x)
            x = self.conv1(x)
            n, c, h, w = x.shape
            x = x.view((n, c*h*w))
            x = self.fc(x)

            return x

    torch.manual_seed(1)
    torch.cuda.manual_seed_all(1)
    np.random.seed(1)
    # random.seed(1)

    torch.cuda.set_device(0)
    net = Net().cuda()
    dummy_input = torch.ones(1, 2, 2, 2).cuda()
    target = torch.ones(1, 100).cuda()
    criterion = nn.MSELoss()
    replace_tuple = (nn.Conv2d, nn.Linear, nn.AvgPool2d)
    net.train()
    optimizer = torch.optim.SGD(net.parameters(), lr=0.01)
    loss = None
    for i in range(200):
        optimizer.zero_grad()
        out = net(dummy_input)
        loss = criterion(out, target)
        if i % 20 == 0:
            print('loss: ', loss)
        loss.backward()
        optimizer.step()
    assert loss < 1, 'training loss error'
    net.eval()
    torch.save(net.state_dict(), 'data.ignore/tool_test.pt')
    linger.wb_analyse('data.ignore/tool_test.pt', 'data.ignore/wb_anylse.log')



from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from ..quant import NormalizeFunction


class NormalizeEmbedding(nn.Embedding):
    def __init__(self, num_embeddings: int, embedding_dim: int, padding_idx: Optional[int] = None,
                 max_norm: Optional[float] = None, norm_type: float = 2., scale_grad_by_freq: bool = False,
                 sparse: bool = False, _weight: Optional[Tensor] = None,  normalize_data=None, normalize_weight=None) -> None:

        assert normalize_data is None or normalize_data > 0, 'normalize value is None or must >0'
        assert normalize_weight is None or normalize_weight > 0, 'normalize value is None or must >0'
        nn.Embedding.__init__(self, num_embeddings, embedding_dim, padding_idx,
                              max_norm, norm_type, scale_grad_by_freq, sparse, _weight)
        self.normalize_data = normalize_data
        self.normalize_weight = normalize_weight

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        normalized_weight = self.weight
        if self.normalize_weight is not None:
            normalized_weight = NormalizeFunction.apply(
                normalized_weight, self.normalize_weight, self.training)

        out = None
        out = F.embedding(input, normalized_weight, self.padding_idx, self.max_norm,
                        self.norm_type, self.scale_grad_by_freq, self.sparse)

        if self.normalize_data is not None:
            out = NormalizeFunction.apply(out, self.normalize_data, self.training, False)
        return out

    def extra_repr(self):
        s = nn.Embedding.extra_repr(self)
        extra_s = ',normalize_data:{normalize_data},normalize_weight:{normalize_weight}}'.format(
            **self.__dict__)
        return s+extra_s


__all__ = ['NormalizeEmbedding']



import itertools
from collections import OrderedDict

import torch
import torch.onnx
from torch.onnx import is_in_onnx_export

from ..config import config
from ..quant import (Quant, normalize_bias_with_config,
                     normalize_weight_with_config)
from ..utils import PlatFormQuant, QuantMode, ScalerBuffer


class ModuleIntConfig():
    def __init__(self, data_bits=8, parameter_bits=8, mode=QuantMode.QValue, o_bits=None):
        self.data_bits = data_bits
        self.parameter_bits = parameter_bits
        self.quant_mode = mode
        self.o_bits = o_bits
        self.quant = Quant()

    @staticmethod
    def state_dict_global(module, destination=None, prefix='', keep_vars=False):
        if destination is None:
            destination = OrderedDict()
            destination._metadata = OrderedDict()
        destination._metadata[prefix[:-1]
                              ] = local_metadata = dict(version=module._version)
        if is_in_onnx_export():
            assert module._buffers['running_x'] > 0, 'invalid running_x <= 0, cannot access param before training, layer prefix is: {}'.format(
                prefix)
            scale_x = ScalerBuffer(module._buffers['scale_x'])
            if module.is_not_from_iqtensor:
                scale_x = ScalerBuffer(module.quant.running_to_scale(ScalerBuffer(
                    module._buffers['running_x']), module.data_bits, mode=module.quant_mode))
                module._buffers['scale_x'].data.fill_(scale_x())
            if module.o_bits is not None:
                scale_o = ScalerBuffer(module.quant.running_to_scale(ScalerBuffer(
                    module._buffers['running_o']), module.o_bits, mode=module.quant_mode))
                module._buffers['scale_o'].data.fill_(scale_o())

        if 'scale_w' in module._buffers and module._parameters['weight'].dtype == torch.float:
            weight_tensor = module._parameters['weight']
            weight_tensor_clamp = None
            bias_tensor_clamp = None
            if hasattr(module, 'clamp_weight'):
                weight_tensor_clamp = normalize_weight_with_config(
                    weight_tensor, module.clamp_weight, False)
            else:
                weight_tensor_clamp = weight_tensor
            weight_tensor.data = weight_tensor_clamp
            if module.bias is not None:
                bias_tensor = module._parameters['bias']
                # bias_temp = None
                if hasattr(module, 'clamp_bias'):
                    bias_tensor_clamp = normalize_bias_with_config(
                        bias_tensor, module.clamp_bias, False)
                else:
                    bias_tensor_clamp = bias_tensor
                bias_tensor.data = bias_tensor_clamp
            if is_in_onnx_export():
                weight_temp, scale_w, _ = module.quant.quant(
                    weight_tensor_clamp, module.parameter_bits, mode=module.quant_mode)
                scale_w = ScalerBuffer(scale_w)
                module._buffers['scale_w'].data.fill_(scale_w())

                if module.parameter_bits <= 8:
                    weight_tensor.data = weight_temp.char()
                    weight_tensor.char()
                elif module.parameter_bits <= 16:
                    weight_tensor.data = weight_temp.short()
                    weight_tensor.short()
                else:
                    weight_tensor.data = weight_temp.int()
                    weight_tensor.int()
                if module.bias is not None:
                    bias_tensor_clamp = module._parameters['bias']
                    if config.PlatFormQuant.platform_quant in (PlatFormQuant.luna_quant,):
                        assert module.quant_mode == QuantMode.QValue, 'luna_quant only support Qvalue and o_bits=None'
                        if module.data_bits + module.parameter_bits <= 16:
                            module._parameters['bias'].data = (
                                bias_tensor_clamp * scale_w * scale_x + 0.5).floor().float().int()
                        else:
                            module._parameters['bias'].data = (
                                bias_tensor_clamp * scale_w * scale_x + 0.5).floor().int()
                        module._parameters['bias'].int()
                    else:
                        assert False, "linger only support luna quant."
        module._save_to_state_dict(destination, prefix, keep_vars)
        for name, module in module._modules.items():
            if module is not None:
                module.state_dict(destination, prefix +
                                  name + '.', keep_vars=keep_vars)
        for hook in module._state_dict_hooks.values():
            hook_result = hook(module, destination, prefix, local_metadata)
            if hook_result is not None:
                destination = hook_result
        return destination

    @staticmethod
    def _load_from_state_dict_global(module, state_dict, prefix, local_metadata, strict,
                                     missing_keys, unexpected_keys, error_msgs):
        allow_missing_keys = ['running_w', 'running_x', 'running_y', 'running_o', 'running_i', 'running_iw', 'running_h', 'running_hw', 'running_io', 'running_ho',
                              'running_i_reverse', 'running_iw_reverse', 'running_h_reverse', 'running_hw_reverse', 'running_io_reverse', 'running_ho_reverse',
                              'scale_w', 'scale_x', 'scale_y', 'scale_o', 'scale_i', 'scale_h', 'scale_iw', 'scale_hw', 'sale_io', 'scale_ho', 'scale_i_reverse', 'scale_h_reverse',
                              'scale_iw_reverse', 'scale_hw_reverse', 'sale_io_reverse', 'scale_ho_reverse', 'running_c', 'scale_c', 'scale_cw', 'min_thresh', 'max_thresh',
                              "running_co", "scale_io", "scale_co", "sigmoid_table", "tanh_table", 'scale_o_reverse', 'scale_io_reverse', 'running_o_reverse',
                              "running_q", "running_k", "running_v", "running_attn", "scale_q", "scale_k", "scale_v", "scale_attn", "running_pos", "scale_pos"]
        local_missing_keys = []
        ModuleIntConfig._load_from_state_dict_global_(module, state_dict, prefix, local_metadata, strict,
                                                      local_missing_keys, unexpected_keys, error_msgs)
        matched = True
        fake_missing_keys = []
        for k_local in local_missing_keys:
            if k_local.replace(prefix, '', 1) not in allow_missing_keys:
                matched = False
                fake_missing_keys.append(k_local)
        if matched:
            local_missing_keys = []
        else:
            local_missing_keys = fake_missing_keys
        missing_keys += local_missing_keys

    @staticmethod
    def _load_from_state_dict_global_(module, state_dict, prefix, local_metadata, strict,
                                      missing_keys, unexpected_keys, error_msgs):
        for hook in module._load_state_dict_pre_hooks.values():
            hook(state_dict, prefix, local_metadata, strict,
                 missing_keys, unexpected_keys, error_msgs)
        local_name_params = itertools.chain(
            module._parameters.items(), module._buffers.items())
        local_state = {k: v.data for k,
                       v in local_name_params if v is not None}
        for name, param in local_state.items():
            key = prefix + name
            if key in state_dict:
                input_param = state_dict[key]

                # Backward compatibility: loading 1-dim tensor from 0.3.* to version 0.4+
                if len(param.shape) == 0 and len(input_param.shape) == 1:
                    input_param = input_param[0]

                if input_param.shape != param.shape:
                    # local shape should match the one in checkpoint
                    error_msgs.append('size mismatch for {}: copying a param with shape {} from checkpoint, '
                                      'the shape in current model is {}.'
                                      .format(key, input_param.shape, param.shape))
                    continue

                if isinstance(input_param, torch.nn.Parameter):
                    input_param = input_param.data
                try:
                    param.copy_(input_param)
                    if input_param.dtype == torch.int32:
                        module._parameters[name] = param.int()
                    elif input_param.dtype == torch.int16:
                        module._parameters[name] = param.short()
                    elif input_param.dtype == torch.int8:
                        module._parameters[name] = param.char()

                except Exception:
                    error_msgs.append('While copying the parameter named "{}", '
                                      'whose dimensions in the model are {} and '
                                      'whose dimensions in the checkpoint are {}.'
                                      .format(key, param.size(), input_param.size()))
            elif strict:
                missing_keys.append(key)
        if strict:
            for key in state_dict.keys():
                if key.startswith(prefix):
                    input_name = key[len(prefix):]
                    input_name = input_name.split('.', 1)[0]
                    if input_name not in module._modules and input_name not in local_state:
                        unexpected_keys.append(key)


__all__ = ['ModuleIntConfig']



import os

import linger
import numpy as np
import torch
import torch.nn as nn
from linger.utils import PlatFormQuant

if not os.path.exists('data.ignore'):
    os.mkdir('data.ignore')


def test_full_training():
    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.transpose = nn.ConvTranspose2d(2, 2, 5, 5, 2, 4, 2, True, 2)
            self.conv1 = nn.Sequential(
                nn.Conv2d(2, 2, kernel_size=3, stride=1,
                          padding=1, bias=False, groups=2),
                nn.BatchNorm2d(2),
                nn.ReLU(),)
            self.conv2 = nn.Sequential(
                nn.Conv2d(2, 2, kernel_size=3, stride=1,
                          padding=1, bias=False, groups=2),
                nn.BatchNorm2d(2),
                nn.ReLU(),)
            self.fc = nn.Linear(392, 100)

        def forward(self, x):
            x = self.transpose(x)
            x = self.conv1(x)
            x = self.conv2(x)
            n, c, h, w = x.shape
            x = x.view((n, c*h*w))
            x = self.fc(x)
            return x

    torch.manual_seed(1)
    torch.cuda.manual_seed_all(1)
    np.random.seed(1)

    torch.cuda.set_device(0)
    net = Net().cuda()
    dummy_input = torch.randn(1, 2, 2, 2).cuda()
    target = torch.ones(1, 100).cuda()
    criterion = nn.MSELoss()
    net.train()
    optimizer = torch.optim.SGD(net.parameters(), lr=0.01)

    # normal training
    for i in range(1000):
        optimizer.zero_grad()
        out = net(dummy_input)
        loss = criterion(out, target)
        if i % 100 == 99:
            print('origin loss: ', loss)
        loss.backward()
        optimizer.step()
    net.eval()
    out_ori = net(dummy_input)
    torch.save(net.state_dict(), 'data.ignore/model.pt.ignore')

    # normal finetune
    net = Net().cuda()
    criterion = nn.MSELoss()
    replace_tuple = (nn.Conv2d, nn.ConvTranspose2d, nn.Linear)
    linger.trace_layers(net, net, dummy_input)
    linger.SetPlatFormQuant(platform_quant=PlatFormQuant.luna_quant)
    # linger.SetCastorBiasInt16(True)
    net = linger.init(net, quant_modules=replace_tuple,
                      mode=linger.QuantMode.QValue)
    net.train()
    net.load_state_dict(torch.load(
        'data.ignore/model.pt.ignore', map_location='cpu'))
    optimizer = torch.optim.SGD(net.parameters(), lr=0.01)
    for i in range(150):
        optimizer.zero_grad()
        out = net(dummy_input)
        loss = criterion(out, target)
        if i % 30 == 29:
            print('loss: ', loss)
        loss.backward()
        optimizer.step()
    net.eval()
    out = net(dummy_input)
    # save int model
    torch.save(net.state_dict(), 'data.ignore/aa.pt')
    out1 = net(dummy_input)
    assert out.sum() == out1.sum()

    # normal testing
    net1 = Net().cuda()
    net1.train()
    linger.trace_layers(net1, net1, dummy_input)
    linger.SetPlatFormQuant(platform_quant=PlatFormQuant.luna_quant)
    # linger.SetCastorBiasInt16(True)
    net1 = linger.init(net1, quant_modules=replace_tuple,
                       mode=linger.QuantMode.QValue)
    net1.eval()
    net1.load_state_dict(torch.load('data.ignore/aa.pt', map_location='cpu'))
    out2 = net1(dummy_input)

    assert out1.sum() == out2.sum()

    with torch.no_grad():
        torch.onnx.export(net, dummy_input, "data.ignore/convtranspose_conv_linear.onnx", export_params=True,
                          opset_version=12, operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    out3 = net(dummy_input)
    assert out2.sum() == out3.sum()



import os

import linger
import numpy as np
import torch
import torch.nn as nn
from linger.utils import PlatFormQuant

if not os.path.exists('data.ignore'):
    os.mkdir('data.ignore')


def test_group_fuse_bn():
    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.transpose = nn.ConvTranspose2d(2, 2, 5, 5, 2, 4, 2, True, 2)
            self.conv = nn.Conv2d(2, 2, kernel_size=3, stride=1,
                                  padding=1, bias=True, groups=2)
            self.bn = nn.BatchNorm2d(2)
            self.fc = nn.Linear(8, 100)

        def forward(self, x):
            x = self.transpose(x)
            x = self.conv(x)
            x = self.bn(x)
            n, c, h, w = x.shape
            x = x.view(n, c*h*w)
            x = self.fc(x)
            return x

    class Net1(nn.Module):
        def __init__(self):
            super(Net1, self).__init__()
            self.transpose = nn.ConvTranspose2d(2, 2, 5, 5, 2, 4, 2, True, 2)
            self.conv = nn.Conv2d(2, 2, kernel_size=3, stride=1,
                                  padding=1, bias=True, groups=2)
            self.bn = nn.BatchNorm2d(2)
            self.fc = nn.Linear(8, 100)

        def forward(self, x):
            x = self.transpose(x)
            x = self.conv(x)
            x = self.bn(x)
            n, c, h, w = x.shape
            x = x.view(n, c*h*w)
            x = self.fc(x)
            return x

    torch.manual_seed(1)
    torch.cuda.manual_seed_all(1)
    np.random.seed(1)
    torch.cuda.set_device(0)
    dummy_input = torch.randn(1, 2, 2, 2).cuda()

    target = torch.ones(1, 100).cuda()
    # criterion = nn.MSELoss()
    replace_tuple = (nn.Conv2d, nn.ConvTranspose2d, nn.Linear)
    net1 = Net().cuda()
    net1.train()
    net1(dummy_input)
    torch.save(net1.state_dict(), 'data.ignore/model.pt.ignore')
    net1.eval()
    out1 = net1(dummy_input)

    net2 = Net1().cuda()
    print(net2)
    linger.trace_layers(net2, net2, dummy_input)
    print(net2)
    linger.SetPlatFormQuant(platform_quant=PlatFormQuant.luna_quant)
    net3 = linger.init(net2, quant_modules=replace_tuple)
    net3.load_state_dict(torch.load('data.ignore/model.pt.ignore'))
    # net3.cuda()
    out3 = net3(dummy_input)
    print(out1.sum() - out3.sum())

    assert out1.sum() - out3.sum() < 1



import os

import linger
import numpy as np
import torch
import torch.nn as nn

if not os.path.exists('data.ignore'):
    os.mkdir('data.ignore')


def test_batchnorm_int_net():
    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.conv = nn.Conv2d(2, 2, kernel_size=2,
                                  stride=1, padding=1, bias=False, groups=2)
            self.bn = nn.BatchNorm2d(2)
            self.relu = nn.ReLU()
            self.conv1 = nn.Conv2d(2, 2, kernel_size=2,
                                   stride=1, padding=1, bias=False, groups=2)
            self.bn1 = nn.BatchNorm2d(2)
            self.relu1 = nn.ReLU()
            self.fc = nn.Linear(32, 100)

        def forward(self, x: dict):
            x = x['input']
            x = self.conv(x)
            x = self.bn(x)
            x = self.relu(x)
            x = self.conv1(x)
            x = self.bn1(x)
            x = self.relu1(x)
            n, c, h, w = x.shape
            x = x.view((n, c*h*w))
            x = self.fc(x)
            return x

    torch.manual_seed(1)
    torch.cuda.manual_seed_all(1)
    np.random.seed(1)

    torch.cuda.set_device(0)
    net = Net().cuda()
    bb = torch.randn(1, 2, 2, 2).cuda()
    aa = {}
    aa['input'] = bb
    aa['input1'] = bb

    target = torch.ones(1, 100).cuda()
    criterion = nn.MSELoss()
    net.train()
    optimizer = torch.optim.SGD(net.parameters(), lr=0.01)

    net = Net().cuda()
    criterion = nn.MSELoss()
    replace_tuple = (nn.Conv2d, nn.ConvTranspose2d, nn.Linear,
                     nn.BatchNorm2d, linger.NormalizeConvBN2d)
    linger.SetPlatFormQuant(platform_quant=linger.PlatFormQuant.luna_quant)
    # linger.FuseBNIntoConv(net, aa)
    linger.trace_layers(net, net, aa, fuse_bn=False)
    net = linger.init(net, quant_modules=replace_tuple,
                      mode=linger.QuantMode.QValue)
    net.train()
    optimizer = torch.optim.SGD(net.parameters(), lr=0.01)
    for i in range(150):
        optimizer.zero_grad()
        out = net(aa)
        loss = criterion(out, target)
        if i % 30 == 29:
            print('loss: ', loss)
        loss.backward()
        optimizer.step()

    net.eval()
    out = net(aa)

    torch.save(net.state_dict(), 'data.ignore/aa.pt')
    out1 = net(aa)
    with torch.no_grad():
        torch.onnx.export(net, aa, "data.ignore/batchnormInt.onnx", export_params=True,
                          opset_version=11, operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    out2 = net(aa)
    assert (out.mean() - out1.mean()
            ) < 0.001, print('out1: {}, out2: {}'.format(out.sum(), out1.sum()))
    assert out.abs().sum() == out2.abs().sum(), 'inconsistant for batchnormint'



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import linger
import numpy as np
import torch
import torch.nn as nn


def test_conv_dropout_linear():
    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.conv = nn.Conv2d(10, 10, kernel_size=3, stride=1,
                                  padding=1, bias=True)
            self.dropout = nn.Dropout(0.5)
            self.fc = nn.Linear(100, 100)

        def forward(self, x):
            x = self.conv(x)
            x = x.squeeze()
            x = x.permute(0, 1, 2)  # permute only support len(x.shape)<=3
            x = self.dropout(x)
            c, h, w = x.shape
            x = x.reshape((c, h*w))
            x = self.fc(x)
            return x
    torch.manual_seed(1)
    torch.cuda.manual_seed_all(1)
    np.random.seed(1)
    # random.seed(1)

    torch.cuda.set_device(0)
    net = Net().cuda()
    dummy_input = torch.randn(1, 10, 10, 10).cuda()
    target = torch.ones(1, 100).cuda()
    criterion = nn.MSELoss()
    replace_tuple = (nn.Conv2d, nn.Linear, nn.AvgPool2d)
    net = linger.init(net)
    net.train()
    optimizer = torch.optim.SGD(net.parameters(), lr=0.01)
    loss = None
    for i in range(200):
        optimizer.zero_grad()
        out = net(dummy_input)
        loss = criterion(out, target)
        if i % 20 == 0:
            print('loss: ', loss)
        loss.backward()
        optimizer.step()
    # assert loss < 1e-12, 'training loss error'
    net.eval()
    torch.save(net.state_dict(), 'data.ignore/conv_permute_linear.pt')
    out1 = net(dummy_input)
    # print(out1)
    with torch.no_grad():
        torch.onnx.export(net, dummy_input, "data.ignore/conv_permute_linear.onnx", export_params=True,
                          opset_version=11, operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    assert abs(out1.mean() - 1) < 1


import logging
import os
from enum import Enum

import numpy as np
import torch

logger = logging.getLogger("linger")
logger.setLevel(logging.INFO)
formatter = logging.Formatter('%(asctime)s  - %(message)s')
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
ch.setFormatter(formatter)
logger.addHandler(ch)


class Singleton(object):
    def __new__(cls, *args, **kw):
        if not hasattr(cls, '_instance'):
            orig = super(Singleton, cls)
            cls._instance = orig.__new__(cls, *args, **kw)
        return cls._instance


class PlatFormQuant(Enum):
    luna_quant = 1


class QuantMode(Enum):
    QValue = 1


class QuantInfo():
    def __init__(self):
        self.data_bits = 8
        self.parameter_bits = 8
        self.output_bits = None
        self.mode = QuantMode.QValue

    def set_data_bits(self, bits):
        self.data_bits = bits

    def set_parameter_bits(self, bits):
        self.parameter_bits = bits

    def set_output_bits(self, bits):
        self.output_bits = bits

    def set_mode(self, mode):
        self.mode = mode


class ClampInfo():
    def __init__(self):
        self.clamp_weight_value = 8
        self.clamp_bias_value = 8
        self.clamp_output_value = None

    def set_clamp_weight_value(self, value):
        self.clamp_weight_value = value

    def set_clamp_bias_value(self, value):
        self.clamp_bias_value = value

    def set_clamp_output_value(self, value):
        self.clamp_output_value = value


def get_device(model):
    device = None
    for parameter in model.parameters():
        device = parameter.device
        break
    assert device is not None, 'paramers of model:%s is empty' % (
        model._get_name())
    return device


def get_max_value(input):
    max_value = -1
    if isinstance(input, list):
        input_tmp = [data.detach() for data in input]
        for data in input_tmp:
            tmp = torch.max(torch.abs(data))
            max_value = max_value if max_value > tmp else tmp
    else:
        input_tmp = input.detach()
        tmp = torch.max(torch.abs(input_tmp))
        max_value = max_value if max_value > tmp else tmp
    return max_value


class Dump():
    @staticmethod
    def dump_file(header, c_name, print_list, file_path):
        if not os.path.exists(file_path):
            os.makedirs(file_path)
        for name, item in print_list:
            if(isinstance(item, torch.Tensor)):
                np.savetxt(os.path.join(file_path, header+c_name+name),
                           item.detach().reshape(-1).cpu().numpy(), fmt="%.6f")


def _unbind(src_tensor):
    dim_0 = src_tensor.size(0)
    nums = dim_0.item() if isinstance(dim_0, torch.Tensor) else dim_0
    sub_tensor_list = [each.squeeze(0)
                       for each in src_tensor.split([1]*nums, dim=0)]
    return sub_tensor_list


def _unbind_packed(packed_tensor, batch_sizes):
    offset = 0
    tensor_list = []
    batch_size_list = [each.item() for each in batch_sizes]
    for batch_len in batch_size_list:
        t = packed_tensor.narrow(0, offset, batch_len)
        offset += batch_len
        tensor_list.append(t)
    return tensor_list, batch_size_list


def _slice(input, start, end):
    if isinstance(input, (tuple)):
        return tuple([each.narrow(0, start, end-start) for each in input])
    else:
        return input.narrow(0, start, end-start)


def hx_slice(input_hidden, cur_hidden, last_batch_size, cur_batch_size):
    if input_hidden is None:  # forward: slice cur_hidden
        assert cur_batch_size < last_batch_size, 'error: forward batch_sizes is not desc order'
        return _slice(cur_hidden, 0, cur_batch_size)
    else:  # reverse: add cur_hidden and sliced input_hidden
        assert cur_batch_size > last_batch_size, 'error: reverse batch_sizes is not asc order'
        slice_hidden = _slice(input_hidden, last_batch_size, cur_batch_size)
        if isinstance(slice_hidden, tuple):
            new_hidden = [torch.cat((cur_e, slice_e), 0)
                          for cur_e, slice_e in zip(cur_hidden, slice_hidden)]
            return tuple(new_hidden)
        return torch.cat((cur_hidden, slice_hidden), 0)


def qshift_round_away_from_zero(x, ):
    x_mask = torch.ones_like(x)
    x_mask[x < 0] = -1
    x_abs = (x.abs() + 0.5).floor()
    x = x_abs * x_mask
    return x


class ScalerBuffer():
    def __init__(self, value):
        if isinstance(value, torch.Tensor):
            value = value.item()
        elif isinstance(value, ScalerBuffer):
            value = value.data
        self.value = np.float32(value)

    def __repr__(self):
        return self.value

    def __str__(self):
        return str(self.value)

    def fill_(self, x):
        if isinstance(x, torch.Tensor):
            x = x.item()
        elif isinstance(x, ScalerBuffer):
            x = x.data
        self.value = np.float32(x)

    def __add__(self, x):
        if isinstance(x, torch.Tensor):
            x = x.item()
        elif isinstance(x, ScalerBuffer):
            x = x.data
        return ScalerBuffer(self.value + np.float32(x))

    def __radd__(self, x):
        return x + self.value

    def add_(self, x):
        if isinstance(x, torch.Tensor):
            x = x.item()
        elif isinstance(x, ScalerBuffer):
            x = x.data
        self.value = self.value + np.float32(x)
        return self

    def __mul__(self, x):
        if isinstance(x, torch.Tensor):
            x = x.item()
        elif isinstance(x, ScalerBuffer):
            x = x.data
        return ScalerBuffer(self.value * np.float32(x))

    def __rmul__(self, x):
        return x * self.value

    def mul_(self, x):
        if isinstance(x, torch.Tensor):
            x = x.item()
        elif isinstance(x, ScalerBuffer):
            x = x.data
        self.value = self.value * np.float32(x)
        return self

    def __truediv__(self, x):
        if isinstance(x, torch.Tensor):
            x = x.item()
        elif isinstance(x, ScalerBuffer):
            x = x.data
        return ScalerBuffer(self.value / np.float32(x))

    def __rtruediv__(self, x):
        return x/self.value

    def __gt__(self, x):
        if isinstance(x, torch.Tensor):
            x = x.item()
        elif isinstance(x, ScalerBuffer):
            x = x.data
        return True if self.value > np.float32(x) else False

    def __eq__(self, x):
        if isinstance(x, torch.Tensor):
            x = x.item()
        elif isinstance(x, ScalerBuffer):
            x = x.data
        return True if self.value == np.float32(x) else False

    def __ne__(self, x):
        if isinstance(x, torch.Tensor):
            x = x.item()
        elif isinstance(x, ScalerBuffer):
            x = x.data
        return True if self.value != np.float32(x) else False

    def __call__(self):
        return self.value

    @property
    def data(self):
        return self.value


__all__ = ['Singleton', 'PlatFormQuant', 'QuantMode', 'QuantInfo', 'ClampInfo', 'get_max_value', 'ClampInfo', 'get_device',
           '_unbind', '_unbind_packed', '_slice', 'hx_slice', 'qshift_round_away_from_zero', 'ScalerBuffer', 'logger']



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import torch
from linger import NormalizeLinear


def test_normalize_linear_foward():
    module = NormalizeLinear(512, 128, True, normalize_data=4,
                             normalize_weight=4, normalize_bias=4)
    module.weight.data.fill_(8)
    module.bias.data.fill_(8)
    input = 8 * torch.randn(2, 512)
    assert (input < 8).any()
    assert (input > 4).any()
    m = module(input)
    assert (m < 4.0001).all()
